//
// Sparc Engine
// OH2NLT / Juha Niinikoski, 26.03.2009
//
//
// Timer & PWM unit usage

// TMR1 is used for 1ms timer tick
// TMR2 tacho out pulse
// TMR3 coil saver
// TMR4 last cranc pulse duration
// TMR5 used for delays (separate module)

// option to count both cranc sensor edges
#define BOTH_EDGES

#include "cop_ignition.h"					// board & processor specific definitions

#define TICK_PERIOD (FCY/1000)				// 1ms tick
#define COIL_SAVER_CYCLE	542				// Clock = FCY/8, 542,53us / step

// engine constants
#ifdef BOTH_EDGES
#define P_CRANC_REV		200					// pulses / Cranc revolution
#else
#define P_CRANC_REV		100					// pulses / Cranc revolution, count both edges
#endif

#define P_CAM_REV		(P_CRANC_REV*2)		// pulses / Cam revolution

#ifdef BOTH_EDGES
//#define TACHO_GATE		300				// Tacho pulse count gate time(ms), 1 RPM resolution
#define TACHO_GATE		30					// Tacho pulse count gate time(ms), 10 RPM resolution
#else
//#define TACHO_GATE		600				// Tacho pulse count gate time(ms), 1 RPM resolution
#define TACHO_GATE		60					// Tacho pulse count gate time(ms), 10 RPM resolution
#endif

#define TACHO_OUT_DIV	(P_CRANC_REV/2)		// Generate Tacho out pulse every N:tc cranc pulse
#define TACHO_OUT_PERIOD (FCY/500)			// TACHO & ECM out pulse 2ms

// Base advance pulse counts
#define BASE_ADV_1 P_CAM_REV				// max count + 1, or "0
#define BASE_ADV_2 (P_CAM_REV/4)
#define BASE_ADV_3 ((P_CAM_REV/4)*2)
#define BASE_ADV_4 ((P_CAM_REV/4)*3)


// Tachometer
volatile int tacho_gate;					// Tachometer gating timer
static volatile unsigned int tacho_ctr;		// IRQ counter
volatile unsigned int tacho;				// Tacho output RPM
volatile int tacho_ready;					// measurement ready

// Diag & datalog
volatile int cam_sensor_diag;				// cranc pulses saved when we get CAM trigger
int cam_sensor_diag_ready;					// diag data available

int datalog_pulses;							// Cylinder #x pulses after CAM trigger
int datalog_pulses_ready;					// datalog data available

extern unsigned long cam_pulses;			// cam sensor test
extern unsigned long cranc_ctr;				// cranc pulses / 400 counter
int cranc_400_ctr;

// ignition timing
static volatile int cam_ctr;				// Cam counter
static volatile int cam_sync_req;			// flag for CAM sync request 

volatile unsigned int cranc_pulse_duration;	// one cranc sensor pulse duration(T4 units, 4,340us/bit)

volatile int	advance_teeth;				// ignition advance set value, # of teeth, must be > 0
volatile int	dwell_teeth;				// dwell set value, # of teeth
volatile int	total_teeth;				// advance+dwell
int irq_temp;								// temp variable for irq

volatile int	coil_1_start;				// start loading pulse count
volatile int	coil_2_start;
volatile int	coil_3_start;
volatile int	coil_4_start;

volatile int	coil_1_ignite;				// ignite pulse count
volatile int	coil_2_ignite;
volatile int	coil_3_ignite;
volatile int	coil_4_ignite;

volatile int	coil_saver;					// coil saver timer set value(ms)

// global variables
volatile unsigned int ms_ctr;				// free running mod 65536 ms counter
volatile unsigned int ms_10_ctr;			// free running 65536 10ms counter
volatile unsigned int seconds;				// free running mod 65536 second counter

int tacho_test;								// flag for tacho test
unsigned char tacho_test_val;				// test value(ms)

// local module variables
static unsigned int ms_counter;				// for counting seconds
static unsigned int ms_10_counter;			// 10x1ms tasks
static unsigned char tacho_test_ctr;		// tacho test counter(ms)


// init timer & Interrupt system
void init_timers_int_system(void)
	{
	cam_sync_req = 0;

	ms_counter = 1000;

	tacho_gate = TACHO_GATE;
	tacho_ctr = 0;
	tacho_ready = 0;

	cam_ctr = 0;
	cranc_400_ctr = 400;

	coil_1_ignite = 0;						// set base advance coil load start points
	coil_2_ignite = BASE_ADV_2;
	coil_3_ignite = BASE_ADV_3;
	coil_4_ignite = BASE_ADV_4;

	coil_1_start = coil_1_ignite - 2;		// at very slow speed coil saver routine triggers the spark
	coil_2_start = coil_2_ignite - 2;
	coil_3_start = coil_3_ignite - 2;
	coil_4_start = coil_4_ignite - 2;

	coil_saver = (16000000 / COIL_SAVER_CYCLE) - 1;		// set some value for coil saver timer, about 16ms

// TMR1 = 1ms tick
	PR1 = TICK_PERIOD - 1;		// TMR1 period register, 1000Hz/1ms
	T1CONbits.TON = 1;			// start TMR1
	_T1IE = 1;					// enable TMR1 tick irq

// TMR2 = Tacho out pulse
	TMR2 = 0;
	PR2 = TACHO_OUT_PERIOD-1;
	T2CONbits.TON = 0;			// do not start timer yet
	_T2IE = 1;

// TMR3 = coil saver variable tick period
	T3CONbits.TCKPS = 1;		// set prescaler to 8
	TMR3 = 0;
	PR3 = coil_saver;			// coil saver period
	T3CONbits.TON = 1;			// start coil saver timer
	_T3IE = 1;					// enable TMR1 tick irq

// TMR4 = cranc pulse duration meter
	T4CONbits.TCKPS = 0x1;		// set prescaler to 8		// 542ns / bit
//	T4CONbits.TCKPS = 0x2;		// set prescaler to 64		// 4,340us / bit
	TMR4 = 0;
	_T4IE = 0;					// no interrupts
	_T4IF = 0;
	T4CONbits.TON = 1;

// Cranc input
	cam_ctr = P_CAM_REV + 10;	// hold / wait for first Cam pulse reset

	_INT3IE = 1;				// enable INT3 IRQ, rising edge sensitive
#ifdef BOTH_EDGES
	_INT4EP = 1;				// select falling edge
	_INT4IE = 1;				// enable INT4 IRQ, falling edge sensitive
#endif

// Cam input
	_INT1IE = 1;				// enable INT1 IRQ, rising edge sensitive

// Datalog event trigger, Cyl#1 sparc
//	_INT2EP = 1;				// select falling edge
//	_INT2IE = 1;				// enable INT2 IRQ, falling edge sensitive

	}


// IRQ code
void sparc_code(void);

#if 0
// Data logger
// Cyl#x sparc, falling input INT2
void __attribute__((interrupt)) _INT2Interrupt(void)
	{
	_INT2IF = 0;
	datalog_pulses = cam_ctr;
	datalog_pulses_ready = 1;
	}
#endif

// Cranc rising input INT3
void __attribute__((interrupt)) _INT3Interrupt(void)
	{
	_INT3IF = 0;

#ifdef RISING_CAM_SYNC
// do cam sync on next rising edge of cranc sensor
	if(cam_sync_req != 0)					// cam sync required
		{
		cam_sync_req = 0;					// reset request flag
		cam_sensor_diag = cam_ctr;			// save sensor diag data
		cam_ctr = 0;
		cam_sensor_diag_ready = 1;
		cam_pulses++;
		}
#endif

	sparc_code();
	}

// Cranc falling input INT4
void __attribute__((interrupt)) _INT4Interrupt(void)
	{
	_INT4IF = 0;

#ifdef FALLING_CAM_SYNC
// do cam sync on next falling edge of cranc sensor
	if(cam_sync_req != 0)					// cam sync required
		{
		cam_sync_req = 0;					// reset request flag
		cam_sensor_diag = cam_ctr;			// save sensor diag data
		cam_ctr = 0;
		cam_sensor_diag_ready = 1;
		cam_pulses++;
		}
#endif

	sparc_code();
	}

// CAM rising input INT1
void __attribute__((interrupt)) _INT1Interrupt(void)
	{
	_INT1IF = 0;

	cam_sync_req = 1;						// set cam sync request flag

#if 0
	cam_sensor_diag = cam_ctr;				// save sensor diag data
	cam_ctr = 0;
	cam_sensor_diag_ready = 1;
	cam_pulses++;
#endif
	}

// Cranc pulse IRQ process, called from rising & falling edge IRQs
void sparc_code(void)
	{
//	TP1 = 1;	// test about 6.6 to 7us

	tacho_ctr++;							// increment tachometer pulse counter

	if(cam_ctr < (P_CAM_REV + 10))			// hold if no Cam pulse when ten over work round
		cam_ctr++;

#if 0
	if( cam_ctr == P_CAM_REV)				// mod(P_REV) ctr, test without cam pulse
		{
		cam_ctr = 0;
		}
#endif

	if(_T4IF)								// check if TMR4 overflow
		{
		_T4IF = 0;
		cranc_pulse_duration = 65535;		// set max count
		}
	else
		{
		cranc_pulse_duration = TMR4;		// save measurement
		}
	TMR4 = 0;								// clear counter

// start loading coil & start ecm/tacho output pulse
	if(cam_ctr == coil_1_start)		// check coil 1 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_1 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

	if(cam_ctr == coil_2_start)	 	// check coil 2 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_2 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

	if(cam_ctr == coil_3_start)		// check coil 3 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_3 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

	if(cam_ctr == coil_4_start)		// check coil 4 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_4 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

// ignite & calculate new values
	if(cam_ctr == coil_1_ignite)	// check coil 1 match
		{
		COIL_1 = 0;					// release spark # 1

		coil_3_ignite = BASE_ADV_3 - advance_teeth;		// calculate new values for # 3
		coil_3_start = coil_3_ignite - dwell_teeth;
		}

	if(cam_ctr == coil_2_ignite)	// check coil 2 match
		{
		COIL_2 = 0;					// release spark # 2

		coil_4_ignite = BASE_ADV_4 - advance_teeth;		// calculate new values for # 4
		coil_4_start = coil_4_ignite - dwell_teeth;
		}

	if(cam_ctr == coil_3_ignite)	// check coil 3 match
		{
		COIL_3 = 0;					// release spark # 3

		coil_1_ignite = BASE_ADV_1 - advance_teeth;		// calculate new values for # 1
		coil_1_start = coil_1_ignite - dwell_teeth;

		}

	if(cam_ctr == coil_4_ignite)	// check coil 4 match
		{
		COIL_4 = 0;					// release spark # 4

		coil_2_ignite = BASE_ADV_2 - advance_teeth;		// calculate new values for # 2
		irq_temp = coil_2_ignite - dwell_teeth;			// check negative overflow for #1
		if(irq_temp < 0)
			coil_2_start = P_CAM_REV + irq_temp;		// overflow, go previous round
		else
			coil_2_start = irq_temp;					// everything is ok
		}

	cranc_400_ctr--;
	if(cranc_400_ctr == 0)
		{
		cranc_400_ctr = 400;
		cranc_ctr++;
		}
//	TP1 = 0;	// test
	}



// Tacho & ECM output pulse end
void __attribute__((interrupt)) _T2Interrupt(void)
	{
//	TP80 = 1;	// test about 475ns

	_T2IF = 0;
	T2CONbits.TON = 0;			// stop TMR2
	TMR2 = 0;					// clear for next round

	ECM_OUT = 0;				// zero outputs
	TACHO_OUT = 0;

//	TP80 = 0;	// test
	}

// Coil saver timer
// also triggers spark at very low engine speeds
void __attribute__((interrupt)) _T3Interrupt(void)
	{
	_T3IF = 0;
	COIL_PORT = COIL_PORT & ~COIL_MASK;		// clear all coil outputs, ignite if loaded
	}


// 1ms Tick timer
void __attribute__((interrupt)) _T1Interrupt(void)
	{
	int x;
	_T1IF = 0;

// tachometer
	tacho_gate--;
	if(tacho_gate == 0)					// count ready
		{
		tacho_gate = TACHO_GATE;		// reload gate timer
		tacho = tacho_ctr;				// save result
		tacho_ctr = 0;					// reset counter
		tacho_ready = 1;				// set ready flag
		}


// general timing
	ms_ctr++;							// global free running 1ms counter

	ms_counter--;						// count for seconds
	if(ms_counter == 0)
		{
		ms_counter = 1000;
		seconds++;						//global free running 1s counter
		}

// 10ms task scheduler
	switch(ms_10_counter)				// 10 ms tasks
		{
		case 0:							// led flasher

		break;

		case 1:

		break;

		case 2:

		break;

		case 3:

		break;

		case 4:

		break;

		case 5:

		break;

		case 6:

		break;

		case 7:

		break;

		case 8:

		break;

		case 9: default:
			ms_10_ctr++;				// run free running 65536 10ms counter
		break;
		}

	if(ms_10_counter >= 9)
		ms_10_counter = 0;
	else
		ms_10_counter++;

	if(tacho_test != 0)							// generate tacho test values if test mode selected
		{
		tacho_test_ctr--;
		if(tacho_test_ctr == 0)
			{
			tacho_test_ctr = tacho_test_val;	// reload timer
			TACHO_OUT = 1;						// start tacho out pulse
			T2CONbits.TON = 1;					// start TMR2
			}
		}
	}
