//
// Test Timers & PWM routines
// OH2NLT / Juha Niinikoski, 26.03.2009
//
//
// Timer & PWM unit usage

// TMR1 is used for 1ms timer tick
// TMR2 tacho out pulse
// TMR3 coil saver
// TMR4 last cranc pulse duration
// TMR5 used for delays (separate module)

#define TICK_PERIOD (FCY/1000)				// 1ms tick
#define COIL_SAVER_CYCLE	542				// Clock = FCY/8, 542,53us / step

#include "cop_ignition.h"					// board & processor specific definitions


// engine constants
#define P_CRANC_REV		100						// pulses / Cranc revolution
#define P_CAM_REV		(P_CRANC_REV*2)			// pulses / Cam revolution
#define TACHO_GATE		600						// Tacho pulse count gate time(ms)

#define TACHO_OUT_DIV	(P_REV/2)			// Generate Tacho out pulse every N:tc cranc pulse
#define TACHO_OUT_PERIOD (FCY/500)			// TACHO & ECM out pulse 2ms


// Tachometer
volatile int tacho_gate;					// Tachometer gating timer
static volatile unsigned int tacho_ctr;		// IRQ counter
volatile unsigned int tacho;				// Tacho output RPM
volatile int tacho_ready;					// measurement ready

// ignition timing
static volatile int cam_ctr;				// Cam counter 

volatile unsigned int cranc_pulse_duration;	// one cranc sensor pulse duration(T4 units, 4,340us/bit)

volatile int	coil_1_start;				// start loading pulse count
volatile int	coil_2_start;
volatile int	coil_3_start;
volatile int	coil_4_start;

volatile int	coil_1_ignite;				// ignite pulse count
volatile int	coil_2_ignite;
volatile int	coil_3_ignite;
volatile int	coil_4_ignite;

volatile int	coil_saver;					// coil saver timer set value(ms)

// global variables
volatile unsigned int ms_ctr;				// free running mod 65536 ms counter
volatile unsigned int ms_10_ctr;			// free running 65536 10ms counter
volatile unsigned int seconds;				// free running mod 65536 second counter

// local module variables
static unsigned int ms_counter;			// for counting seconds
static unsigned int ms_10_counter;		// 10x1ms tasks



// init timer & Interrupt system
void init_timers_int_system(void)
	{
	ms_counter = 1000;

	tacho_gate = TACHO_GATE;
	tacho_ctr = 0;
	tacho_ready = 0;

	cam_ctr = 0;

	coil_1_start = 0;						// set base advance coil load start points
	coil_2_start = (P_CAM_REV/4);
	coil_3_start = ((P_CAM_REV/4)*2);
	coil_4_start = ((P_CAM_REV/4)*3);

	coil_1_ignite = coil_1_start + 2;		// at very slow speed coil saver routine triggers the spark
	coil_2_ignite = coil_2_start + 2;
	coil_3_ignite = coil_3_start + 2;
	coil_4_ignite = coil_4_start + 2;

	coil_saver = (16000000 / COIL_SAVER_CYCLE) - 1;		// set some value for coil saver timer, about 16ms

// TMR1 = 1ms tick
	PR1 = TICK_PERIOD - 1;		// TMR1 period register, 1000Hz/1ms
	T1CONbits.TON = 1;			// start TMR1
	_T1IE = 1;					// enable TMR1 tick irq

// TMR2 = Tacho out pulse
	TMR2 = 0;
	PR2 = TACHO_OUT_PERIOD-1;
	T2CONbits.TON = 0;			// do not start timer yet
	_T2IE = 1;

// TMR3 = coil saver variable tick period
	T3CONbits.TCKPS = 1;		// set prescaler to 8
	TMR3 = 0;
	PR3 = coil_saver;			// coil saver period
	T3CONbits.TON = 1;			// start coil saver timer
	_T3IE = 1;					// enable TMR1 tick irq

// TMR4 = cranc pulse duration meter
	T4CONbits.TCKPS = 0x1;		// set prescaler to 8		// 542ns / bit
//	T4CONbits.TCKPS = 0x2;		// set prescaler to 64		// 4,340us / bit
	TMR4 = 0;
	_T4IE = 0;					// no interrupts
	_T4IF = 0;
	T4CONbits.TON = 1;

// Cranc input
	_INT3IE = 1;				// enable INT3 IRQ
	}


// IRQ code

// Cranc rising input INT3
void __attribute__((interrupt)) _INT3Interrupt(void)
	{
	TP1 = 1;	// test about 4.4 to 5us

	_INT3IF = 0;
	tacho_ctr++;							// increment tachometer pulse counter

	cam_ctr++;
	if( cam_ctr == P_CAM_REV)				// mod(P_REV) ctr
		cam_ctr = 0;

	if(_T4IF)								// check if TMR4 overflow
		{
		_T4IF = 0;
		cranc_pulse_duration = 65535;		// set max count
		}
	else
		{
		cranc_pulse_duration = TMR4;		// save measurement
		}
	TMR4 = 0;								// clear counter

// start loading coil & start ecm/tacho output pulse
	if(cam_ctr == coil_1_start)		// check coil 1 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_1 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

	if(cam_ctr == coil_2_start)	 	// check coil 2 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_2 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

	if(cam_ctr == coil_3_start)		// check coil 3 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_3 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

	if(cam_ctr == coil_4_start)		// check coil 4 match
		{
		TMR3 = 0;					// reset coil saver
		COIL_4 = 1;					// start loading

		ECM_OUT = 1;				// set ECM & Tacho outputs
		TACHO_OUT = 1;
		T2CONbits.TON = 1;			// start TMR2
		}

// ignite
	if(cam_ctr == coil_1_ignite)	// check coil 1 match
		{
		COIL_1 = 0;					// start loading
		}

	if(cam_ctr == coil_2_ignite)	// check coil 2 match
		{
		COIL_2 = 0;					// start loading
		}

	if(cam_ctr == coil_3_ignite)	// check coil 3 match
		{
		COIL_3 = 0;					// start loading
		}

	if(cam_ctr == coil_4_ignite)	// check coil 4 match
		{
		COIL_4 = 0;					// start loading
		}

	TP1 = 0;	// test
	}



// Tacho & ECM output pulse end
void __attribute__((interrupt)) _T2Interrupt(void)
	{
	TP80 = 1;	// test about 475ns

	_T2IF = 0;
	T2CONbits.TON = 0;			// stop TMR2
	TMR2 = 0;					// clear for next round

	ECM_OUT = 0;				// zero outputs
	TACHO_OUT = 0;

	TP80 = 0;	// test
	}

// Coil saver timer
// also triggers spark at very low engine speeds
void __attribute__((interrupt)) _T3Interrupt(void)
	{
	_T3IF = 0;
	COIL_PORT = COIL_PORT & ~COIL_MASK;		// clear all coil outputs, ignite if loaded
	}


// 1ms Tick timer
void __attribute__((interrupt)) _T1Interrupt(void)
	{
	int x;
	_T1IF = 0;

// tachometer
	tacho_gate--;
	if(tacho_gate == 0)					// count ready
		{
		tacho_gate = TACHO_GATE;		// reload gate timer
		tacho = tacho_ctr;				// save result
		tacho_ctr = 0;					// reset counter
		tacho_ready = 1;				// set ready flag
		}


// general timing
	ms_ctr++;							// global free running 1ms counter

	ms_counter--;						// count for seconds
	if(ms_counter == 0)
		{
		ms_counter = 1000;
		seconds++;						//global free running 1s counter
		}

// 10ms task scheduler
	switch(ms_10_counter)				// 10 ms tasks
		{
		case 0:							// led flasher

		break;

		case 1:

		break;

		case 2:

		break;

		case 3:

		break;

		case 4:

		break;

		case 5:

		break;

		case 6:

		break;

		case 7:

		break;

		case 8:

		break;

		case 9: default:
			ms_10_ctr++;				// run free running 65536 10ms counter
		break;
		}

	if(ms_10_counter >= 9)
		ms_10_counter = 0;
	else
		ms_10_counter++;

	}
