//
// Simple UART2 interface for dsPIC30F6014
// Juha Niinikoski, OH2NLT 27.10.2008
//

#include "cop_ignition.h"	
#include <stdio.h>

//#define BAUD 9600
//#define FCY  1000000	// 4MHz XTAL
//#define FCY  8000000	// 4MHz XTAL x8 PLL = 32MHz
//#define FCY  16000000	// 4MHz XTAL x16 PLL = 64MHz
//#define FCY  29491200	// 7,3728MHz XTAL x16 PLL = 117,9648MHz


// Baud Rate divisors for 30 MHz osc =  7,5 MHz FCY 
/* actual baud rate divisors
[0] 1200	389,625
[1] 2400	194,3125
[2] 4800	96,65625
[3] 9600	47,828125
[4] 19200	23,4140625
[5] 38400	11,20703125
[6] 57600	7,138020833
[7] 115200	3,069010417
*/


// Baud Rate divisors for 29,4912 MHz osc =  7,3728 MHz FCY 
/* actual baud rate divisors
[0] 1200	383
[1] 2400	191
[2] 4800	95
[3] 9600	47
[4] 19200	23
[5] 38400	11
[6] 57600	7
[7] 115200	3
*/



/* Module variables & definitions */

static volatile unsigned char rx_buffer[256];      // UART mod(256) RX queue, must be exactly 256 long
static volatile unsigned char rx_buf_out_idx = 0;  // buffer output index, pointer behind in idx
static volatile unsigned char rx_buf_in_idx = 1;   // buffer input index, pointer always "leading"

// 7,5MHz
//const unsigned int baud_rates[8] = {390, 194, 97, 48, 23, 11, 7, 3};

// 7,3728MHz
static const unsigned int baud_rates[8] = {383, 191, 95, 47, 23, 11, 7, 3};


// UART1 IRQ service
// queue character
void __attribute__((interrupt)) _U2RXInterrupt(void)  // put receiced character to RX queue
	{
  	unsigned char rx;

  	rx = U2RXREG;                        			// data read
  	if((unsigned char)(rx_buf_in_idx + 1) != rx_buf_out_idx) 		// check overrun
  		{
  		rx_buffer[rx_buf_in_idx++] = rx;    		// place just received character in queue
    	}
  	// else reject character
// check & handle possible errors
	if(U2STAbits.OERR == 1)							// OERR is blocking uart, should not happen but..
		U2STAbits.OERR = 0;							// reset overrun error if occured

// this is not automatic in 30F6014 UART
	IFS1bits.U2RXIF = 0;							// reset irq flag
	}


// UART1 init
void InitUSART2(unsigned long baud)
{

 U2BRG = ((FCY/16)/baud) - 1;   		// calculate & set baud rate
// U2BRG = baud_rates[3];					// set intial baud rate to 9600

 U2MODE = 0x0000;
 U2STA = 0x0000;						// IRQ on every character
 IFS1bits.U2RXIF = 0;					// clear possible pending irq
 IEC1bits.U2RXIE = 1;					// enable RX IRQs
 U2MODEbits.UARTEN = 1;					// enable uart
 U2STAbits.UTXEN = 1;					// enable transmitter, must be done after UARTEN
}

// set baud rate from the table
void SetUSART2baud(int baud_sel)
	{
	if(baud_sel > 7)
		baud_sel = 7;					// stay in the table

	 U2BRG = baud_rates[baud_sel];		// set baud rate
	}

// print character
void putch2(unsigned char c)
	{
	while(U2STAbits.UTXBF == 1)		// wait here if tx fifo full
		{
		}
	U2TXREG = (int)c &0x00FF;	// Tx reg is 16-bit long
	}

// print string
void putst2(unsigned char *str)
	{
	while((*str)!=0)
		{
		putch2(*str);
		str++;
		}
	} 

void putchhex2(unsigned char c)
	{
	unsigned char temp;
	temp=c;

	c=(c >> 4);
	if (c<10) c+=48; else c+=55;
	putch2(c);
	c=temp;
	c=(c & 0x0F);
	if (c<10) c+=48; else c+=55;
	putch2(c);
	}


// check if key perssed, data available from receive buffer
// Not pressed = 0, Pressed = number of characters
// IRQ based RX queue check
int kbhit2(void)
	{
//	if(U1STAbits.OERR == 1)								// OERR is blocking uart, should not happen but..
//		U1STAbits.OERR = 0;								// reset overrun error if occured
  	return (unsigned char)((rx_buf_in_idx  - ((unsigned char)rx_buf_out_idx + 1)));   	// return with number of available characters
	}

// clear serial port, reset input buffer
void ClearUSART2queue(void)
	{
	rx_buf_out_idx = 0;  								// buffer output index
	rx_buf_in_idx = 1;   								// buffer input index
	}


// get char from RX queue, IRQ RX system
unsigned char getch2(void)
	{
 	while((unsigned char)(rx_buf_out_idx + 1) == rx_buf_in_idx);   	// wait data if queue is empty 
  	return rx_buffer[++rx_buf_out_idx];             	// return with data
	}


// receive character with echo
unsigned char getche2(void)
	{
	unsigned char c;
	putch2(c = getch2());
	return c;
	}


// some debug I/O functions
// Get unsigned long

unsigned long getlong2(void)
	{
	unsigned long val;
	char tmp;
	val = 0;
	
	do
		{
		tmp = getche2() & 0x7F;
		putch1(tmp);
		if(tmp >= '0' && tmp <= '9')
			val = 10 * val + (tmp - '0');
		}
	while (tmp != 0x0D);
	
	return(val);
	}

// Special output functions

static unsigned char disp_buffer[11];		// display buffer for long number printouts

/* Convert binary number to ASCII string */
/* 10 digits are converted */

static void bin2ascii(unsigned long luku)
	{
char x, idx;

	idx = 9;
	for(x=0; x<10; x++)				// do 10 digits
		{
		disp_buffer[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	disp_buffer[10] = 0;			// string terminator
	}

void putlong2( unsigned long x)
	{
	bin2ascii(x);					// partial convert long number to string

	putch2(disp_buffer[0]);		// MSB

	putch2(disp_buffer[1]);		// 
	putch2(disp_buffer[2]);		// 
	putch2(disp_buffer[3]);		//

	putch2(disp_buffer[4]);		//
	putch2(disp_buffer[5]);		//
	putch2(disp_buffer[6]);		//

	putch2(disp_buffer[7]);		//
	putch2(disp_buffer[8]);		//
	putch2(disp_buffer[9]);		// LSB
	}
